#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <getopt.h>
#include <time.h>

#include <netcdf.h>


#define NDIMS		3
#define NLAT		94
#define NLON		192
#define NTIMEMAX	366	// Max 366 days in a year
#define NAME_LAT	"lat"
#define NAME_LON	"lon"
#define NAME_TIME	"time"
#define NAME_AIR	"air"

#define NELEM(x)	(sizeof(x) / sizeof(*(x)))

#define ERR(e)	do { fprintf(stderr, "Error: %s\n", nc_strerror(e)); exit(1); } while(0)

char *dir_i = "./noaa_air_2m";
char *dir_o = "./noaa_air_2m_img";
char *name_base = "air.2m.gauss";

static float latangles[NLAT] = {
	 88.5420, 86.6531, 84.7532, 82.8508, 80.9473, 79.0435, 77.1394, 75.2351, 
	 73.3307, 71.4262, 69.5217, 67.6171, 65.7125, 63.8079, 61.9033, 59.9986, 
	 58.0939, 56.1893, 54.2846, 52.3799, 50.4752, 48.5705, 46.6658, 44.7611, 
	 42.8564, 40.9517, 39.0470, 37.1422, 35.2375, 33.3328, 31.4281, 29.5234, 
	 27.6186, 25.7139, 23.8092, 21.9044, 19.9997, 18.0950, 16.1902, 14.2855, 
	 12.3808, 10.47604, 8.57131, 6.66657, 4.76184, 2.8571, 0.952368, 
	 -0.952368, -2.8571,  -4.76184, -6.66657, -8.57131, -10.47604, -12.3808,
	-14.2855, -16.1902, -18.0950, -19.9997, -21.9044, -23.8092, -25.7139, -27.6186,
	-29.5234, -31.4281, -33.3328, -35.2375, -37.1422, -39.0470, -40.9517, -42.8564,
	-44.7611, -46.6658, -48.5705, -50.4752, -52.3799, -54.2846, -56.1893, -58.0939,
	-59.9986, -61.9033, -63.8079, -65.7125, -67.6171, -69.5217, -71.4262, -73.3307,
	-75.2351, -77.1394, -79.0435, -80.9473, -82.8508, -84.7532, -86.6531, -88.5420
};

#define NPOINTS		(NLON*NLAT)
#define NBUFDAYS	(NTIMEMAX*2)
typedef struct { float d[NPOINTS]; } daydata_t;
static daydata_t *daydata;
static size_t ddn;	// # of datapoints in dataset
static size_t ddhead;	// Point for addition
static size_t ddtail;	// Point for removal

// Buffer to hold two years of data so we always can have a 365 day windows
// within the data for average calculation
static void fillbuffer(daydata_t *data, size_t n)
{
	if(ddhead + n > NBUFDAYS) {
		memcpy(&daydata[ddhead], data, (NBUFDAYS - ddhead)*sizeof(*data));
		memcpy(&daydata[0], data + NBUFDAYS - ddhead, (ddhead + n - NBUFDAYS)*sizeof(*data));
		ddhead = ddhead + n - NBUFDAYS;
	} else {
		memcpy(&daydata[ddhead], data, n*sizeof(*data));
		ddhead += n;
	}
	ddn += n;
}

// Read one year's dataset
static size_t readfile(const char *fname, daydata_t **buf)
{
	int retval;
	int ncid;
	// int varid_lat, varid_lon;
	int varid_time, varid_air;

	if(!fname)
		return 0;

	// Non-existent file is not an error...
	struct stat sb;
	if(stat(fname, &sb) < 0) {
		if(ENOENT == errno)
			return 0;
		fprintf(stderr, "Stat error on %s: %s\n", fname, strerror(errno));
		exit(1);
	}

	/* Open the file. */
	if ((retval = nc_open(fname, NC_NOWRITE, &ncid)))
		ERR(retval);

	/* Get the varids of the latitude and longitude coordinate variables. */
	if ((retval = nc_inq_varid(ncid, NAME_AIR, &varid_air)))
		ERR(retval);
	if ((retval = nc_inq_varid(ncid, NAME_TIME, &varid_time)))
		ERR(retval);
/*
	if ((retval = nc_inq_varid(ncid, NAME_LAT, &varid_lat)))
		ERR(retval);
	if ((retval = nc_inq_varid(ncid, NAME_LON, &varid_lon)))
		ERR(retval);
*/

	// Number of days
	size_t ntime = 0;
	if((retval = nc_inq_dim(ncid, varid_time, NULL, &ntime)))
		ERR(retval);

	*buf = calloc(ntime, sizeof(**buf));

	/* Read data */
	if((retval = nc_get_var_float(ncid, varid_air, (*buf)->d))) {
		ERR(retval);
	}
	/* Close the file. */
	if ((retval = nc_close(ncid)))
		ERR(retval);
	return ntime;
}

static size_t count_days(int syear, int tyear)
{
	daydata_t *data;
	char fname[1024];
	size_t sum = 0;

	for(int i = syear; i <= tyear; i++) {
		snprintf(fname, sizeof(fname), "%s/%s.%d.nc", dir_i, name_base, i);
		size_t ntime = readfile(fname, &data);
		if(!ntime) {
			fprintf(stderr, "Count days breaks at year %d, file not found?\n", i);
			break;
		}
		sum += ntime;
		free(data);
	}
	return sum;
}

static float *calc_year_avg(int syear, int years)
{
	float *avg;
	int navg = 0;
	daydata_t *data;
	char fname[1024];

	avg = calloc(NLAT*NLON, sizeof(*avg));

	// Sum years adding each day
	for(int i = syear; i < syear+years; i++) {
		snprintf(fname, sizeof(fname), "%s/%s.%d.nc", dir_i, name_base, i);
		size_t ntime = readfile(fname, &data);
		if(!ntime) {
			fprintf(stderr, "Average breaks at year %d, file not found?\n", i);
			break;
		}
		for(size_t t = 0; t < ntime; t++) {
			navg++;
			for(size_t o = 0; o < NLON*NLAT; o++) {
				avg[o] += data[t].d[o];
			}
		}
		free(data);
	}

	// Divide by # of days processed for average and go to degrees Celcius
	for(size_t o = 0; o < NLON*NLAT; o++) {
		avg[o] = avg[o] / navg - 273.15;
	}

	return avg;
}

static void sum_rem_add(daydata_t *avg, daydata_t *rem, daydata_t *add)
{
	if(add && !rem) {
		for(size_t o = 0; o < NLON*NLAT; o++) {
			avg->d[o] += add->d[o];
		}
	} else if(!add && rem) {
		for(size_t o = 0; o < NLON*NLAT; o++) {
			avg->d[o] -= rem->d[o];
		}
	} else if(add && rem) {
		for(size_t o = 0; o < NLON*NLAT; o++) {
			avg->d[o] += add->d[o] - rem->d[o];
		}
	}
}

static const char *nextfile(int *nextyear, int lastyear)
{
	static char fname[1024];
	if(*nextyear > lastyear)
		return NULL;
	snprintf(fname, sizeof(fname), "%s/%s.%d.nc", dir_i, name_base, *nextyear);
	++(*nextyear);
	return fname;
}

static void nextday(daydata_t *sum, int ndays, int *fyear, int tyear)
{
	sum_rem_add(sum, &daydata[ddtail], &daydata[(ddtail+ndays)%NBUFDAYS]);
	ddtail++;
	ddtail %= NBUFDAYS;
	ddn--;
	if(NBUFDAYS - ddn >= NTIMEMAX) {
		daydata_t *data;
		size_t n = readfile(nextfile(fyear, tyear), &data);
		if(n) {
			fillbuffer(data, n);
			free(data);
		}
	}
}
static const char usage_str[] =
	"Usage: noaadata [options]\n"
	"Options:\n"
	"  -h       This message\n"
	"  -b <str> Name base for .nc files (default: air.2m.gauss)\n"
	"  -i <dir> Input directory where .nc files are read (default: ./noaa_air_2m)\n"
	"  -o <dir> Output directory where .ppm files are put (default: ./noaa_air_2m_img)\n"
	"  -f <int> From year (default: 1948)\n"
	"  -t <int> To year (default: 2021)\n"
	"  -s <int> Start year for average (default: 1948)\n"
	"  -y <int> # of years in average (default: 30)\n"
	"  -z <int> # of days in moving average smoothing (default: 365)\n"
	;

int main(int argc, char *argv[])
{
	int optc;
	int lose = 0;
	int syear = 1948;
	int years = 30;
	int fyear = 1948;
	int tyear = 2021;
	int ndays = 365;
	char *eptr;

	while(EOF != (optc = getopt(argc, argv, "hy:s:i:o:f:t:b:z:"))) {
		switch(optc) {
		case 'y':
			years = strtol(optarg, &eptr, 0);
			if(years < 1 || years > 100 || *eptr) {
				fprintf(stderr, "Years -y must be between 1 and 100\n");
				lose++;
			}
			break;
		case 's':
			syear = strtol(optarg, &eptr, 0);
			if(syear < 1948 || syear > 2020 || *eptr) {
				fprintf(stderr, "Start year -s must be between 1948 and 2020\n");
				lose++;
			}
			break;
		case 'f':
			fyear = strtol(optarg, &eptr, 0);
			if(fyear < 1800 || fyear > 2100 || *eptr) {
				fprintf(stderr, "From year -f must be between 1800 and 2100 (yeah, right)\n");
				lose++;
			}
			break;
		case 't':
			tyear = strtol(optarg, &eptr, 0);
			if(tyear < 1800 || tyear > 2100 || *eptr) {
				fprintf(stderr, "To year -t must be between 1800 and 2100 (yeah, right)\n");
				lose++;
			}
			break;
		case 'i':
			dir_i = strdup(optarg);
			if(!dir_i) {
				fprintf(stderr, "Cannot copy input directory");
				lose++;
			}
			break;
		case 'o':
			dir_o = strdup(optarg);
			if(!dir_o) {
				fprintf(stderr, "Cannot copy output directory");
				lose++;
			}
			break;
		case 'b':
			name_base = strdup(optarg);
			if(!name_base) {
				fprintf(stderr, "Cannot copy base name");
				lose++;
			}
			break;
		case 'z':
			ndays = strtol(optarg, &eptr, 0);
			if(ndays < 1 || ndays > 5000 || *eptr) {
				fprintf(stderr, "# of days in moving average must be between 1 and 5000 in -z\n");
				lose++;
			}
			break;
		case 'h':
			fprintf(stderr, "%s", usage_str);
			lose++;
			break;
		default:
			lose++;
			break;
		}
	}

	if(lose)
		return 1;

	double latavgweight = 0.0;
	for(unsigned u = 0; u < NELEM(latangles); u++) {
		latavgweight += latangles[u] = cos(latangles[u] / 180.0 * M_PI);
	}

	float *yearavg = calc_year_avg(syear, years);
	size_t totaldays = count_days(fyear, tyear);

	daydata = calloc(NBUFDAYS, sizeof(*daydata));
	
	// Read the first years
	while(NBUFDAYS - ddn >= NTIMEMAX) {
		daydata_t *data;
		size_t n = readfile(nextfile(&fyear, tyear), &data);
		fillbuffer(data, n);
		free(data);
	}

	daydata_t *sum = calloc(1, sizeof(*sum));
	daydata_t *daydiff = calloc(1, sizeof(*daydiff));

	// Calculate year sum
	for(int i = 0; i < ndays; i++) {
		sum_rem_add(sum, NULL, &daydata[i]);
	}
	ddtail = 0;

	//int fon = 0;

	float minmini = FLT_MAX;
	float maxmaxi = -FLT_MAX;

	// HACK: so that we will start at 01-Jan-1950
	nextday(sum, ndays, &fyear, tyear);

	printf("P6\n%zu %zu\n255\n", (size_t)NLAT, totaldays - 365);
	size_t nlines = 0;

	while(ddn >= ndays) {
		// Daily 365 day average difference from 30 year average
		float mini = FLT_MAX;
		float maxi = -FLT_MAX;
		for(size_t o = 0; o < NLON*NLAT; o++) {
			daydiff->d[o] =  (sum->d[o] / (float)ndays - 273.15) - yearavg[o];
		}
		//fprintf(stderr, "%f %f\n", mini, maxi);


#define ZEROBAND	0.1
		mini = -7.5 + ZEROBAND;
		maxi = +7.5 - ZEROBAND;

		double tempavgsum = 0.0;
		for(size_t a = 0; a < NLAT; a++) {
			float v = 0.0;
			for(size_t o = 0; o < NLON; o++) {
				v += daydiff->d[a * NLON + ((o+NLON/2+2)%NLON)];
			}

			v /= NLON;
			tempavgsum += v * latangles[a];
			if(v < minmini) minmini = v;
			if(v > maxmaxi) maxmaxi = v;

			uint8_t r, g, b;
			if(fabsf(v) <= ZEROBAND) {
				r = g = b = 255;
			} else {
				if(v < mini) v = mini;
				if(v > maxi) v = maxi;
				if(v > 0) {
					r = 255;
					g = b = 255 - (int)((v - ZEROBAND)/(maxi - ZEROBAND) * 255.0);
				} else {
					b = 255;
					g = r = 255 - (int)((v + ZEROBAND)/(mini + ZEROBAND) * 255.0);
				}
			}
			fputc(r, stdout);
			fputc(g, stdout);
			fputc(b, stdout);
		}
		nlines++;
		//fprintf(stderr, "%lf\n", tempavgsum / latavgweight);

		nextday(sum, ndays, &fyear, tyear);
	}

	fprintf(stderr, "nlines %zu\n", nlines);
	fprintf(stderr, "%+.3f %+.3f\n", minmini, maxmaxi);

	return 0;
}
